﻿using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using GameEngine.Commands;
using GameEngine.Commands.PlayerCommands;
using TestHarness.Exceptions;
using TestHarness.Properties;
using TestHarness.Util;

namespace TestHarness.TestHarnesses.Bot
{
    public abstract class BotRunner
    {
        protected readonly BotHarness ParentHarness;
        protected TimeSpan MaxRunTime;
        private bool _errorLogged = false;

        protected BotRunner(BotHarness parentHarness)
        {
            ParentHarness = parentHarness;

            MaxRunTime = TimeSpan.FromSeconds(Settings.Default.MaxBotRuntimeSeconds);
        }

        protected string AddAdditionalRunArgs(string currentArgs)
        {
            var runArgs = ParentHarness.BotMeta.RunArgs;

            return String.IsNullOrEmpty(runArgs) ? currentArgs : String.Format("{0} {1}", currentArgs, runArgs);
        }

        public void CalibrateBot()
        {
            var stopWatch = new Stopwatch();
            stopWatch.Start();
            RunCalibrationTest();
            stopWatch.Stop();

            MaxRunTime = MaxRunTime.Add(stopWatch.Elapsed);

            ParentHarness.Logger.LogInfo(String.Format("Bot calibration complete and can run an additional {0}ms due to environment startup", stopWatch.ElapsedMilliseconds));
        }

        public void RunBot()
        {
            using (var handler = CreateProcessHandler())
            {
                double runTime = 0;
                try
                {
                    handler.LimitExecutionTime = ParentHarness.EnforceTimeLimit;
                    handler.ProcessToRun.OutputDataReceived += (sender, args) => ParentHarness.Logger.LogInfo("Output from bot: " + args.Data);
                    handler.ProcessToRun.ErrorDataReceived += (sender, args) => ParentHarness.Logger.LogException("Output from bot: " + args.Data);

                    if (ParentHarness.HaltOnError)
                    {
                        handler.ProcessToRun.ErrorDataReceived += (sender, args) =>
                        {
                            _errorLogged = _errorLogged || !String.IsNullOrEmpty(args.Data);
                        };
                    }

                    ParentHarness.Logger.LogDebug(String.Format("Executing bot with following commands {0} {1}", handler.ProcessToRun.StartInfo.FileName, handler.ProcessToRun.StartInfo.Arguments));

                    handler.RunProcess();
                    Thread.Sleep(4000);
                    runTime = (handler.ProcessToRun.ExitTime - handler.ProcessToRun.StartTime).TotalMilliseconds;

                    ParentHarness.Logger.LogInfo("Your bots total execution time was " + TimeSpan.FromMilliseconds(runTime));

                    if (ParentHarness.HaltOnError && _errorLogged)
                    {
                        Console.WriteLine(String.Format("Bot {0} encountered an error, press any key to continue", ParentHarness.PlayerEntity.Key));
                        Console.ReadKey();
                    }
                }
                catch (Exception ex)
                {
                    ParentHarness.Logger.LogException("Failure while executing bot " + handler.ProcessToRun.StartInfo.FileName + " " + handler.ProcessToRun.StartInfo.Arguments, ex);
                }

                if (ParentHarness.EnforceTimeLimit && runTime >= MaxRunTime.TotalMilliseconds)
                {
                    ParentHarness.Logger.LogInfo("Your bot exceeded the maximum execution time");
                    throw new TimeLimitExceededException("Time limit exceeded by " + (MaxRunTime.Subtract(TimeSpan.FromMilliseconds(runTime))));
                }
            }
        }

        public ICommand GetBotCommand()
        {
            try
            {
                var commandCode = GetBotCommandFromFile();

                switch (commandCode)
                {
                    case 1:
                        return new MovementCommand(MovementCommand.Direction.Up);
                    case 2:
                        return new MovementCommand(MovementCommand.Direction.Left);
                    case 3:
                        return new MovementCommand(MovementCommand.Direction.Right);
                    case 4:
                        return new MovementCommand(MovementCommand.Direction.Down);
                    case 5:
                        return new PlaceBombCommand();
                    case 6:
                        return new TriggerBombCommand();
                    default:
                        return new DoNothingCommand();
                }
            }
            catch (Exception ex)
            {
                ParentHarness.Logger.LogException("Failed to read move file from bot", ex);
                return new DoNothingCommand();
            }
        }

        private int GetBotCommandFromFile()
        {
            var workDir = ParentHarness.CurrentWorkingDirectory;
            var moveFile = Path.Combine(workDir, Settings.Default.BotMoveFileName);

            if (!File.Exists(moveFile))
                return 0;

            var code = File.ReadAllText(moveFile).FirstOrDefault();

            return Char.IsNumber(code) ? (int)Char.GetNumericValue(code) : 0;
        }

        protected abstract ProcessHandler CreateProcessHandler();
        protected abstract void RunCalibrationTest();
    }
}
